=begin

Copyright abandoned 2011 Brian Concannon, placed in the public domain by Brian Concannon 

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name :          Face2Face.rb 1.04
Description :   Joins faces to create cables/wires/ducts/pipes/conduits/extrusions in Sketchup
Author :        Brian Concannon, brian_concannon@yahoo.com
Usage :         1. Install into the plugins directory (or subfolder - supports organizer.rb) 
                2. Select any two faces that seem far enough apart.
                3. Select from the Sketchup menu, "Plugins" then "Face2Face"
                4. The circular faces are extruded from one face to the other along an
                   automatically generated path
Date :      2011-12-10
History:
    1.04    (2011-12-10)    Moved and added menu choices to Plugins menu
                            Added selecting of an edge to indicate the 'from' face
                            Added selecting of nearby entities to reverse direction of extrusion from faces
    1.03    (2011-12-08)    Changed 'too close together' logic to improve false messages/incompletions
                            Also allow noncircular faces to be extruded
                            Also better convergence control to reduce iteration count
                            Also changed name from wiretool to Face2Face tool.
    1.02    (2011-11-24)    Using 90 edges per circle, gives very nice joints.  48 did, too.
                            Also removed 'too close' checking, just letting the process fail instead, 
                            which gives more capability for close faces that can work.
                            Note:tried for a long time to allow user to pick faces after selecting this
                            tool, but was not able to deal with rotating tool and drag tool etc.
    1.01    (2011-10-15)    Made code compatible with Organizer.rb
                            Modified the header information to match the template as discussed 
                            on the SketchUcation forum.
                            Added model.start_operation and @m_model.commit_operation calls,
                            to make the final product a single undo/redo.
    1.00    (2011-09-15)    First version, sent to smustard.com for review
=end
=begin
Algorithm: see document Face2Face.pdf
Variable name prefixes to show intended usage of each variable:

  bln_ boolean 1 = true 0-false otherwise unknown.  Note Ruby uses 0xFFFF = true.
                       bls_ set of booleans
  c_   curve/arc       cs_  set of curves/arcs
  e_   entity          es_  set of entities
  ed_  edge            eds_ set of edges  
  f_   face            fs_ vset of faces
  fi_  file
  i_   integer scalar  is_  set of integer scalars
  l_   line
  lo_  loop            los_ set of loops
  m_   model
  p_   point           ps_  set of points
  ph_  pick helper
  pl_  plane
  s_   floating point scalar
                       ss_  set of floating point scalars
  sel_ selection
  tr_  transformation
  t_   text
  v_   vector
  vtx_ vertex          vts_ set of vertices
  @@   class variable
=end
require 'sketchup.rb'
include Math
module MF2F
#-----------------------------------------------------------------------------
# Define the tool class for creating face extrusions along a new path
#-----------------------------------------------------------------------------
class Face2Face
#-----------------------------------------------------------------------------
  FAILED = 0
  FAILURE = 0
  SUCCESS = 1
  MB_YES = 6
  MB_NO = 7
  $i_face_count = 0
  $bln_animate = 0
  $bln_show_construction = 0
  $bln_show_points = 0
  $s_extr_rad_multiplier = 1
  $s_bend_radius_multiplier = 6
  $bln_label_faces = 0
  $s_stub_length_multiplier = 2
  $bln_reverse_followpath = 0
  $bln_reverse_stub1 = 0
  $bln_reverse_stub2 = 0
#-----------------------------------------------------------------------------
def btcAddText(t_text,p_xy)
  if @bln_addtext != 0 
    t_ = @m_model.entities.add_text t_text, p_xy
    return t_
  end
  return nil
end
#-----------------------------------------------------------------------------
def btcGiveUpControl
#nodebugplease
puts 'giving up control'
  @m_model.commit_operation
  @m_model.select_tool nil #tells Sketchup to deselect our tool
  $i_face_count = 0
end
#-----------------------------------------------------------------------------
def btcRescuePrint(t_)
#nodebugplease
   puts 'starting rescueprint'
  tt_ = 'RESCUE FROM ' + t_
  print tt_ 
  UI.messagebox(tt_)
end
#-----------------------------------------------------------------------------
def btcDelay(s_seconds)
  s_tstart = Time.now
  s_tdiff = 0
  while s_tdiff < s_seconds
    s_tdiff = Time.now - s_tstart
  end                           
end
#-----------------------------------------------------------------------------    
def reset
#nodebugplease
  @p_pts = []
  Sketchup::set_status_text 'F2F tool reset'
  @drawn = false
end
#-----------------------------------------------------------------------------

def btcMakev_unitnormal
begin
  i_return = self.btcCalculateFaceCentroid
  if i_return != SUCCESS
    return FAILED
  end
  @v_unitnormal = @e_myentity.normal
  return SUCCESS
rescue
  btcRescuePrint 'btcMakev_unitnormal'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcCalculateFaceCentroid
begin
  f_myface = @e_myentity   
  vts_myvertices = f_myface.vertices
  s_mysumx = 0
  s_mysumy = 0
  s_mysumz = 0
  i_mypoint = 0
  p_mypt = Geom::Point3d.new 0, 0, 0

  vts_myvertices.each {|i|       
    i_mypoint = i_mypoint+1
    s_myx = i.position.x
    s_myy = i.position.y
    s_myz = i.position.z  
    p_mypt = p_mypt.set! s_myx, s_myy, s_myz

    s_mysumx = s_mysumx+s_myx
    s_mysumy = s_mysumy+s_myy
    s_mysumz = s_mysumz+s_myz

    }            
  if i_mypoint < 1 
    UI.messagebox('Face has no points')
    return FAILED
  end
  s_mycentx = s_mysumx/i_mypoint   #centroid of face if circular
  s_mycenty = s_mysumy/i_mypoint
  s_mycentz = s_mysumz/i_mypoint
  p_mycentroid = Geom::Point3d.new  0,0,0
  p_mycentroid.x = s_mycentx
  p_mycentroid.y = s_mycenty
  p_mycentroid.z = s_mycentz
  s_myradius = p_mycentroid.distance p_mypt
  @p_cent = p_mycentroid
  @s_radius = s_myradius
  return SUCCESS
rescue
  btcRescuePrint 'btcCalculateFaceCentroid'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcIsAboveFace(p_n,pl_planen,v_unitnormaln)
begin
  p_ntoplanen = p_n.project_to_plane pl_planen
  v_droppedn = Geom::Vector3d.new p_n.x-p_ntoplanen.x,p_n.y-p_ntoplanen.y,p_n.z-p_ntoplanen.z

  v_droppedn = v_droppedn.reverse
  s_direction = v_droppedn.dot v_unitnormaln
  if s_direction>0    #can't figure out why this is not <0 instead of >0
    @bln_usen = 0
  else
    @bln_usen = 1
  end
  return SUCCESS
rescue
  btcRescuePrint 'btcIsAboveFace'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcCreateOneArc(i_1or2)
begin
  if i_1or2 == 1
    p_centn = @p_cent1offset
    p_bestnOf576 = @p_best1of576
    p_6dcentcirclen = @p_6dcentcircle1
    v_normn_abxad = @v_6dcircleaxis1_abxad  
  else
    p_centn = @p_cent2offset
    p_bestnOf576 = @p_best2of576
    p_6dcentcirclen = @p_6dcentcircle2
    v_normn_abxad = @v_6dcircleaxis2_abxad
  end
  s_dx = p_centn.x-p_6dcentcirclen.x
  s_dy = p_centn.y-p_6dcentcirclen.y
  s_dz = p_centn.z-p_6dcentcirclen.z
  v_gf = Geom::Vector3d.new s_dx,s_dy,s_dz 
  v_fg = v_gf.reverse
  s_dx = p_bestnOf576.x-p_6dcentcirclen.x
  s_dy = p_bestnOf576.y-p_6dcentcirclen.y
  s_dz = p_bestnOf576.z-p_6dcentcirclen.z
  v_gp = Geom::Vector3d.new s_dx,s_dy,s_dz 
  s_angle = v_gf.angle_between v_gp
  if s_angle < 0
    UI.messagebox('angle<0')  #should never happen, but can be = 0.
  end
  i_segs = @i_circle_edgecount*s_angle/6.28
  s_a1 = 0
  s_a2 = s_a1+s_angle
  if i_segs>0
    eds_arc = Sketchup.active_model.entities.add_arc p_6dcentcirclen, v_gf, v_normn_abxad, @s_len6r, s_a1,   s_a2,    i_segs
  else
    eds_arc = nil
  end
  if i_1or2 == 1
    @eds_arc1 = eds_arc
  else
    @eds_arc2 = eds_arc
  end
  return SUCCESS
rescue
  btcRescuePrint 'btcCreateOneArc'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcComputeAngleErrors
begin
    s_dot = @v_test.dot @v_seg12
    if s_dot < 0
      @v_seg12 = @v_seg12.reverse  
    end 
    @s_ang1 = @v_test.angle_between @v_seg12
    s_dot = @v_test.dot @v_seg34
    if s_dot < 0
      @v_seg34 = @v_seg34.reverse  
    end 
    @s_ang2 = @v_test.angle_between @v_seg34
    @s_angsum = @s_ang1.abs + @s_ang2.abs    
    return SUCCESS 
rescue 
  btcRescuePrint 'btcComputeAngleErrors'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcFindPointsAboveFace
begin
  i_found1count = 0
  i_found2count = 0
  i_found3count = 0
  i_found4count = 0
  @i_edgeidx1 = 0
  while @i_edgeidx1<@i_edgecount1
    e = @eds_circle1[@i_edgeidx1]
    @p_1 = e.start.position
    i_return = btcIsAboveFace(@p_1,@pl_plane1offset,@v_unitnormal1)
    if i_return != SUCCESS
      return FAILED
    end
    s_distfromcenter1 = @p_1.distance @p_f1
    if s_distfromcenter1 > @s_len6r * @s_6r_criterion
      @bln_usen = 0
    end
    if @bln_usen == 1
      i_found1count = i_found1count+1
    end
    @bls_usep1[@i_edgeidx1] = @bln_usen
    @p_2 = e.end.position
    i_return = btcIsAboveFace(@p_2,@pl_plane1offset,@v_unitnormal1)
    if i_return != SUCCESS
      return FAILED
    end
    s_distfromcenter1 = @p_2.distance @p_f1
    if s_distfromcenter1 > @s_len6r * @s_6r_criterion
      @bln_usen = 0
    end
    if @bln_usen == 1
      i_found2count = i_found2count+1
    end
    @bls_usep2[@i_edgeidx1] = @bln_usen
    @i_edgeidx1 = @i_edgeidx1+1
  end
  if @bls_usep1[0]==1              #don't use same points twice
      if @bls_usep2[@i_edgecount1-1]==1
      @bls_usep1[@i_edgecount1-1] = 0
      @bls_usep2[@i_edgecount1-1] = 0
    end
  end

  if @bls_usep2[0]==1
      if @bls_usep2[@i_edgecount1-1]==1
      @bls_usep1[@i_edgecount1-1] = 0
      @bls_usep2[@i_edgecount1-1] = 0
    end
  end
#---------------------------------------------------------------------
  if i_found1count == 0
    if i_found2count == 0
      t_ = 'LOGIC ERROR FPAF1 NO POINTS ABOVE OFFSET FACE 1'
      UI.messagebox(t_)
      return FAILED
    end
  end
#---------------------------------------------------------------------
  @i_edgeidx2 = 0
  while @i_edgeidx2<@i_edgecount2
    e2 = @eds_circle2[@i_edgeidx2]
    @p_3 = e2.start.position
    i_return = btcIsAboveFace(@p_3,@pl_plane2offset,@v_unitnormal2)
    if i_return != SUCCESS
      return FAILED
    end
    s_distfromcenter2 = @p_3.distance @p_f2
    if s_distfromcenter2 > @s_len6r * @s_6r_criterion
      @bln_usen = 0
    end
    if @bln_usen == 1
      i_found3count = i_found3count+1
    end
    @bls_usep3[@i_edgeidx2] = @bln_usen
    @p_4 = e2.end.position
    i_return = btcIsAboveFace(@p_4,@pl_plane2offset,@v_unitnormal2)
    if i_return != SUCCESS
      return FAILED
    end
    s_distfromcenter2 = @p_4.distance @p_f2
    if s_distfromcenter2 > @s_len6r * @s_6r_criterion
      @bln_usen = 0
    end
    if @bln_usen == 1
      i_found4count = i_found4count+1
    end
    @bls_usep4[@i_edgeidx2] = @bln_usen
    @i_edgeidx2 = @i_edgeidx2+1
  end
  if @bls_usep2[0]==1              #don't use same points twice
      if @bls_usep3[@i_edgecount2-1]==1

      @bls_usep3[@i_edgecount2-1] = 0
      @bls_usep4[@i_edgecount2-1] = 0
      t_ = 'DEBUGGING-WILL NOT USE LAST SEGMENT OF CIRCLE 2'            
        end
     end
  if @bls_usep4[0]==1
      if @bls_usep4[@i_edgecount2-1]==1

      @bls_usep3[@i_edgecount2-1] = 0
      @bls_usep4[@i_edgecount2-1] = 0
      t_ = 'DEBUGGING-WILL NOT USE LAST SEGMENT OF CIRCLE 2'            
        end
     end
#---------------------------------------------------------------------
  if i_found3count == 0
    if i_found4count == 0
      t_ = 'LOGIC ERROR FPAF2 NO POINTS ABOVE OFFSET FACE 2'

      UI.messagebox(t_)
            return FAILED
    end
  end
#---------------------------------------------------------------------
  return SUCCESS
rescue
  btcRescuePrint 'btcFindPointsAboveFace'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcCalculateOOPError
begin
    i_calc_corr = @is_calccorr[@i_adjust]   
    if i_calc_corr == 0
      @s_outofplaneerror = 0
      return SUCCESS
    end
    e = @eds_circle1[@i_bestindex1]
    e2 = @eds_circle2[@i_bestindex2]
    @p_1 = e.start.position
    @p_2 = e.end.position
    @p_3 = e2.start.position
    @p_4 = e2.end.position
#   @s_angsum = @s_angsumBestOf576
    @p_best1 = @p_best1of576
    @p_best2 = @p_best2of576
    if i_calc_corr == 1
      @v_tangent = Geom::Vector3d.new @p_best2.x-@p_best1.x,@p_best2.y-@p_best1.y,@p_best2.z-@p_best1.z
      p_proj = @p_best2.project_to_plane @pl_6dcircle1
      v_toproj = Geom::Vector3d.new p_proj.x-@p_best1.x,p_proj.y-@p_best1.y,p_proj.z-@p_best1.z
      v_perp = p_proj-@p_best2
      s_dot = v_perp.dot @v_6dcircleaxis1_abxad
    end
    if i_calc_corr == 2
      @v_tangent = Geom::Vector3d.new @p_best1.x-@p_best2.x,@p_best1.y-@p_best2.y,@p_best1.z-@p_best2.z
      p_proj = @p_best1.project_to_plane @pl_6dcircle2
      v_toproj = Geom::Vector3d.new p_proj.x-@p_best2.x,p_proj.y-@p_best2.y,p_proj.z-@p_best2.z
      v_perp = p_proj-@p_best1
      s_dot = v_perp.dot @v_6dcircleaxis2_abxad
    end
    @s_outofplaneerror = @v_tangent.angle_between v_toproj
    if @s_outofplaneerror > 90.degrees
      @s_outofplaneerror = 180.degrees - @s_outofplaneerror  
    end
    if s_dot > 0
      @s_outofplaneerror = -1*@s_outofplaneerror  
    end  
    return SUCCESS
rescue
  btcRescuePrint 'btcCalculateOOPError'
  return FAILED  
end
end
#-----------------------------------------------------------------------------
def btcMakeExtrusion
begin
#  @bln_startAtFace1 = 1
  @s_len6r = $s_bend_radius_multiplier * @s_radius
  @s_clearance = @s_radius * 2 * $s_stub_length_multiplier
  i_return = self.btcGetBestNormals
  if i_return != SUCCESS 
    @m_model.commit_operation
    return FAILED
  end
  @p_best1of576 = [0 ,0 ,0]
  @p_best2of576 = [0 ,0 ,0]
  i_return = 10
  i_return = self.btcCreateBothArcs
  if i_return >= 99
       #straight extrusion, already followed?
    i_return -= 100
    if i_return == SUCCESS 
      if $bln_show_construction == 0
        @m_model.erase_entities @ed_tangent
      end
    end
    return i_return
  end 
  if i_return != SUCCESS 
    return FAILED
  end
  if @p_best1of576==@p_best2of576
    UI.messagebox('tangent line endpoints cannot be the same')
    @m_model.commit_operation
    return FAILURE
  end
  if @s_ang1_best_save > 0.1 
    UI.messagebox('Faces are too close together')
    @m_model.commit_operation
    return FAILURE
  end
  if @s_ang2_best_save > 0.1 
    UI.messagebox('Faces  are too close together')
    @m_model.commit_operation
    return FAILED
  end
  @ed_tangent = Sketchup.active_model.entities.add_line @p_best1of576, @p_best2of576
begin
  s_lentan = @ed_tangent.length
rescue
  s_lentan = -1
end
  if s_lentan < 0
    return FAILED
  end
  Sketchup::set_status_text 'Creating path', SB_VCB_LABEL
# Create extrusion path for followme tool
  i_return = self.btcConstructContinuousPath
  if i_return != SUCCESS 
    return FAILED
  end
  @m_model.commit_operation  #allow user to ctrl-z once without erasing the path
  if Sketchup.version[0,1].to_i >= 7
    @m_model.start_operation("draw surfaces",true)
  else
    @m_model.start_operation "draw surfaces"
  end
  i_return = @f_face1.followme @eds_follow  #use followme tool to
                                 #create the outer surface of the extrusion

  if i_return == nil
    UI.messagebox('followme failed')
    return FAILED
  end    
  @i_nttcount_after = Sketchup.active_model.entities.count
  Sketchup::set_status_text 'Done', SB_VCB_LABEL
  $i_face_count = 0
  return SUCCESS
rescue Exception  >= t_exception
  btcRescuePrint 'EXCEPTION IN btcMakeExtrusion '
  puts 't_exception = '+t_exception.to_s
  @m_model.commit_operation
rescue
  btcRescuePrint 'btcMakeExtrusion '
  @m_model.commit_operation
  return 2
end
end
#-----------------------------------------------------------------------------
def btcStoreBestSoFar
begin
  if @s_angsum < @s_angsumBestOf576
    @s_angsumBestOf576 = @s_angsum
    @p_best1of576 = @p_best1
    @p_best2of576 = @p_best2
    @i_bestindex1 = @i_edgeidx1
    @i_bestindex2 = @i_edgeidx2
    @s_ang1_best_save=@s_ang1
    @s_ang2_best_save=@s_ang2
    @i_bestcnt += 1
    if $bln_animate == 1
      if @s_angsum < @s_bestofAllTries
        @s_bestofAllTries=@s_angsum
        if @ed_tangentx != nil
          @m_model.entities.erase_entities @ed_tangentx
        end
        t_ = @p_best1of576.to_s
        t_ = @p_best2of576.to_s
        @ed_tangentx = Sketchup.active_model.entities.add_line @p_best1of576, @p_best2of576
        t_ = @i_bestcnt.to_s + ' ' + @s_angsum.to_s + ' ' + @s_tinner.to_s
        Sketchup::set_status_text t_, SB_VCB_VALUE
        bln_refreshed_view = @m_model.active_view.refresh 
 #       self.btcDelay 0.167
      end    
    end  
  end
  return SUCCESS
rescue
  btcRescuePrint 'btcStoreBestSoFar'
  return FAILED
end
end
#-----------------------------------------------------------------------------
#
#  Middle iteration tries all likely combinations of segments,
#  one on circle 1 and one on circle 2
#
#-----------------------------------------------------------------
def btcMiddleIteration
begin
# Algorithm step 11- create line tangent to circle1 and circle2, sort of parallel to linea1a2. This is done by
#    selecting circle1 and circle2 segments having slopes nearest the slope of 
#    the tangent lines,
  @i_edgecount1 = @eds_circle1.size
  @i_edgecount2 = @eds_circle2.size
  @p_1 = @p_spare #create any old point
  @p_2 = @p_spare #create any old point
  @p_3 = @p_spare #create any 2 points
  @p_4 = @p_spare
#2222222222222222222222222222222222222222222222222
#
# This is where n legitimate points for circles 1 and 2 are assigned "use/do not use" flags
#
  i_return = self.btcFindPointsAboveFace
  if i_return != SUCCESS
    return FAILED
  end   
#2222222222222222222222222222222222222222222222222
  @i_edgeidx1 = 0
  @s_angsumBestOf576 = 99.2
  @p_best1of576 = [0 ,0 ,0]
  @p_best2of576 = [0 ,0 ,0]
  @ps1_point1 = []
  @ps1_point2 = []
#-----------------------------------------------------------------
#
#  Here is the start of the middle iteration loop
#
#-----------------------------------------------------------------
  while @i_edgeidx1<@i_edgecount1
    @ps2_point1 = []
    @ps2_point2 = []
    @bln_usep1 = @bls_usep1[@i_edgeidx1]
    @bln_usep2 = @bls_usep2[@i_edgeidx1]
#.....................................................................
    if @bln_usep1+@bln_usep2>0  #proceed if one or the other is true
      e = @eds_circle1[@i_edgeidx1]
      @p_1 = e.start.position
      @p_2 = e.end.position
      s_12x = @p_2.x-@p_1.x
      s_12y = @p_2.y-@p_1.y
      s_12z = @p_2.z-@p_1.z
      @v_seg12 = Geom::Vector3d.new s_12x, s_12y, s_12z    
      @i_edgeidx2 = 0
      @s_angsum = 99
#-----------------------------------------------------------------
#     inner iteration loop, iterates on segments on circle 2
#-----------------------------------------------------------------
      @s_t = Time.new
      i_return = self.btcInnermostIteration
      if i_return != SUCCESS
        return FAILED
      end 
      @s_tinner += Time.new - @s_t       
#-----------------------------------------------------------------
    end  #if @bln_usep1+@bln_usep2>0
#.....................................................................
    @i_edgeidx1 += 1
  end  #while @i_edgeidx1<@i_edgecount1 yes
  
  return SUCCESS    
rescue
  self.btcRescuePrint 'btcMiddleIteration'
  return FAILED
end
end
#-----------------------------------------------------------------
#
#  Innermost iteration evaluates a legitimate segment on circle 2
#
#-----------------------------------------------------------------
def btcInnermostIteration
begin
while @i_edgeidx2 < @i_edgecount2
  @bln_usep3 = @bls_usep3[@i_edgeidx2]
  @bln_usep4 = @bls_usep4[@i_edgeidx2]
  if @bln_usep3 + @bln_usep4>0
    e2 = @eds_circle2[@i_edgeidx2]
    @p_3 = e2.start.position
    @p_4 = e2.end.position
    s_34x = @p_4.x-@p_3.x
    s_34y = @p_4.y-@p_3.y
    s_34z = @p_4.z-@p_3.z
    @v_seg34 = Geom::Vector3d.new s_34x, s_34y, s_34z    
    if @bln_usep1 == 1
      if @bln_usep3 == 1
        s_x = @p_3.x-@p_1.x
        s_y = @p_3.y-@p_1.y
        s_z = @p_3.z-@p_1.z
        @p_best1 = @p_1
        @p_best2 = @p_3
      else
        s_x = @p_4.x-@p_1.x
        s_y = @p_4.y-@p_1.y
        s_z = @p_4.z-@p_1.z
        @p_best1 = @p_1
        @p_best2 = @p_4
      end
    else
      if @bln_usep3 == 1
        s_x = @p_3.x-@p_2.x
        s_y = @p_3.y-@p_2.y
        s_z = @p_3.z-@p_2.z
        @p_best1 = @p_2
        @p_best2 = @p_3
      else
        s_x = @p_4.x-@p_2.x
        s_y = @p_4.y-@p_2.y
        s_z = @p_4.z-@p_2.z
        @p_best1 = @p_2
        @p_best2 = @p_4
      end
    end
    @v_test = Geom::Vector3d.new s_x, s_y, s_z    
#    @v_test = @v_test.normalize    
    i_return = self.btcComputeAngleErrors
    if i_return != SUCCESS
      return FAILED
    end              
  end  #if @bln_usep3+@bln_usep4>0
  i_return = self.btcStoreBestSoFar
  if i_return != SUCCESS
    return FAILED
  end              
#.....................................................................
  @i_edgeidx2 += 1
end  #while @i_edgeidx2
return SUCCESS
rescue
  self.btcRescuePrint 'btcInnermostIteration'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcSetupIterations
begin
# The sequence is 14 + 1 steps - 
# Algorithm step 1 - let point a1 be the center of the circular face1
  @p_a1 = @p_cent1
  @pl_plane1 = [@p_a1, @v_unitnormal1]
  self.btcShowPoint @p_a1,0
  t_cent1 = self.btcAddText ' a1', @p_a1

# Algorithm step 2 - let point b1 be the end of the unit normal to the center of face1
  p_b1 = @p_unitend1
  self.btcShowPoint p_b1,0
  t_cent1 = self.btcAddText ' b1', p_b1

  @v_a1b1 = Geom::Vector3d.new p_b1.x-@p_a1.x,p_b1.y-@p_a1.y,p_b1.z-@p_a1.z
  if @v_a1b1.length < 1.mm
    return nil
    t_ = 'LOGIC ERROR CBA8, length a1b1 < 0.1mm'
    UI.messagebox(t_)
  end
  s_ratio = @s_clearance/@v_a1b1.length
  tr_ = Geom::Transformation.scaling s_ratio 
  @v_a1f1 = @v_a1b1.transform tr_
  if @v_a1f1.length < 1.mm
    t_ = 'LOGIC ERROR CBA7, length a1f1 < 0.1mm'
    UI.messagebox(t_)
    return nil
  end
  @p_f1 = @p_a1 + @v_a1f1
  @p_hingepoint1=@p_f1
  if @s_radius < 1.mm
    t_ = 'LOGIC ERROR CBA6, radius < 0.1mm'
    UI.messagebox(t_)
    return nil
  end
  @pl_plane1offset = [@p_f1, @v_unitnormal1]
  @p_cent1offset = @p_f1
# Algorithm step 3 - let point a2 be the center of the circular face2
  @p_a2 = @p_cent2
  self.btcShowPoint @p_a2,0    
  t_cent1 = self.btcAddText ' a2', @p_a2

  @pl_plane2 = [@p_a2, @v_unitnormal2] 
  @v_a1a2 = Geom::Vector3d.new @p_a2.x-@p_a1.x,@p_a2.y-@p_a1.y,@p_a2.z-@p_a1.z 
  @v_a1a2n = @v_a1a2.normalize
# Algorithm step 4 - let point b2 be the end of the unit normal to the center of face2
  p_b2 = @p_unitend2
  self.btcShowPoint p_b2,0
  t_cent1 = self.btcAddText ' b2', p_b2

  @v_a2b2 = Geom::Vector3d.new p_b2.x-@p_a2.x,p_b2.y-@p_a2.y,p_b2.z-@p_a2.z
  s_ratio = @s_clearance/@v_a2b2.length
  tr_ = Geom::Transformation.scaling s_ratio 
  @v_a2f2 = @v_a2b2.transform tr_
  @p_f2 = @p_a2 + @v_a2f2
  @p_hingepoint2=@p_f2
  @pl_plane2offset = [@p_f2, @v_unitnormal2]
  @p_cent2offset = @p_f2
# Algorithm step 5 - create linea1a2 from a1 to a2
  @ed_edgea1a2 = Sketchup.active_model.entities.add_line(@p_a1, @p_a2)
  @l_linea1a2 = @ed_edgea1a2.line
  @m_model.entities.erase_entities @ed_edgea1a2
  @s_lena1a2 = @p_a1.distance @p_a2
# Algorithm step 6 - create point c on linea1a2 half way between a1 and a2
  @p_c = Geom::Point3d.linear_combination 0.5, @p_a1, 0.5, @p_a2
  self.btcShowPoint @p_c,0
  t_cent1 = self.btcAddText ' c', @p_c

  @i_adjust_limit = 20  #if this is increased, also fix @is_adjust and $is_adjust below.
  @i_adjust = 0
  @is_adjust =  [1,2]*@i_circle_edgecount  #double the size it needs to be
  @is_adjust[0] = 0
                                       #adjust circle at start of loop for these extrusion ends
                                       #extrusion end 1 is near face 1, end 2 is near face 2, we
                                       #never know which face was selected first however.
  @is_calccorr = [2,1]*@i_circle_edgecount   #double 
                                       #calculate correction at end of loop for these extrusion ends
  @s_cumulativeadjustment1 = 0
  @s_cumulativeadjustment2 = 0
  @s_outofplaneerrorb4 = 99
  @ss_ooperrors = [99,99,99,99]
  @s_oopLast4b4 = 99
  @s_oopLast4 = 99
  return SUCCESS    
rescue
  self.btcRescuePrint 'btcSetupIterations'
  return FAILED
end
end
#-----------------------------------------------------------------
#
#  Outermost iteration makes a correction to a 6*r circle based
#  on the out-of-plane error of that circle 
#
#
#-----------------------------------------------------------------
def btcOutermostIteration
begin
  i_return = self.btcSetupIterations
  
  if i_return != SUCCESS
    return FAILED
  end                
  i_show_iteration = 0
#-----------------------------------------------------------------
#
#  Here is the start of the outermost iteration loop
#
#-----------------------------------------------------------------
  while @i_adjust < @i_adjust_limit
    i_fix_which = @is_adjust[@i_adjust]
#-----------------------------------------------------------------
#   make correction to plane of large circle
#-----------------------------------------------------------------
    if i_fix_which == 1
      @s_cumulativeadjustment1 = @s_cumulativeadjustment1 + @s_acceleration_factor * @s_outofplaneerror
    end
    if i_fix_which == 2
      @s_cumulativeadjustment2 = @s_cumulativeadjustment2 + @s_acceleration_factor * @s_outofplaneerror
    end

# Algorithm step 7A - project point c onto plane of face1, this is point d1

    p_d1 = @p_c.project_to_plane @pl_plane1
    s_dx = p_d1.x-@p_a1.x
    s_dy = p_d1.y-@p_a1.y
    s_dz = p_d1.z-@p_a1.z
    if s_dx.abs < 0.01.mm

      if s_dy.abs < 0.01.mm

        if s_dz.abs < 0.01.mm
          i_return = self.btcFollowStraightPath @p_cent1, @p_cent2
          return 100 + i_return
        end
      end
    end
    @v_a1d1 = Geom::Vector3d.new s_dx,s_dy,s_dz
                       #puts '@s_cumulativeadjustment1 ' + @s_cumulativeadjustment1.to_s
    if @s_cumulativeadjustment1 != 0
      tr_rotate1 = Geom::Transformation.rotation @p_a1, @v_unitnormal1, @s_cumulativeadjustment1
      @v_a1d1.transform! tr_rotate1  #all reconstruction of the large circle at end 1 is
                                    #affected by this rotation.
    end
    p_d1 = @p_a1 + @v_a1d1
    self.btcShowPoint p_d1,0
    t_cent1 = self.btcAddText ' d1', p_d1
# Algorithm step 7B - project point c onto face2, this is point d2
    p_d2 = @p_c.project_to_plane @pl_plane2
    @v_a2d2 = Geom::Vector3d.new p_d2.x-@p_a2.x,p_d2.y-@p_a2.y,p_d2.z-@p_a2.z
    if @s_cumulativeadjustment2 != 0
      tr_rotate2 = Geom::Transformation.rotation @p_a2, @v_unitnormal2, @s_cumulativeadjustment2
      @v_a2d2.transform! tr_rotate2
                                    #all reconstruction of the large circle at end 2 is
                                    #affected by this rotation.
    end
    p_d2 = @p_a2 + @v_a2d2
    self.btcShowPoint p_d2,0
    t_cent1 = self.btcAddText ' d2', p_d2
# Algorithm step 8A - add point e1 on line a1d1, 6*radius distance from point a1 
#     e1 is the point just below the arc center g1 for the extrusion bend at face1.
    s_dista1d1 = @p_a1.distance p_d1
    if s_dista1d1 < 1.mm
      i_return = self.btcFollowStraightPath @p_cent1, @p_cent2
      return 7
    end
    s_distance_ratio = @s_len6r/s_dista1d1
    tr_ = Geom::Transformation.scaling s_distance_ratio 
    v_a1e1 = @v_a1d1.transform tr_
    p_e1 = @p_a1 + v_a1e1
    @p_g1 = p_e1 + @v_a1f1
    self.btcShowPoint p_e1,0
    self.btcShowPoint @p_f1,0
    self.btcShowPoint @p_g1,0
    t_cent1 = self.btcAddText ' e1', p_e1
    t_cent1 = self.btcAddText ' f1', @p_f1
    t_cent1 = self.btcAddText ' g1', @p_g1
# Algorithm step 8B - add point e2 on line a2d2, 6*radius distance from point a2 
#     e2 is the point just below the arc center at g2 for the extrusion bend at face2.
    s_dista2d2 = @p_a2.distance p_d2
    s_distance_ratio = @s_len6r/s_dista2d2
    tr_ = Geom::Transformation.scaling s_distance_ratio 
    v_a2e2 = @v_a2d2.transform tr_
    if v_a2e2.length < 1.mm
      t_ = 'LOGIC ERROR CBA9, length a2e2 < 0.1mm'
      UI.messagebox(t_)
      return nil
    end
    p_e2 = @p_a2 + v_a2e2 
    @p_g2 = p_e2 + @v_a2f2
    self.btcShowPoint p_e2,0
    self.btcShowPoint @p_f2,0
    self.btcShowPoint @p_g2,0
    t_cent1 = self.btcAddText ' e2', p_e2
    t_cent1 = self.btcAddText ' f2', @p_f2
    t_cent1 = self.btcAddText ' g2', @p_g2
# Algorithm step 9A - find direction of axis of extrusion bend circle near face1
    @v_6dcircleaxis1_abxad = @v_a1b1.cross @v_a1d1  
# Algorithm step 9B - find direction of axis of extrusion bend circle near face2
    @v_6dcircleaxis2_abxad = @v_a2b2.cross @v_a2d2  
# Algorithm step 10A - create circle1 centered at g1, radius 6*r
#     Circle will be tangent to ab at e.
    @eds_circle1 = Sketchup.active_model.entities.add_circle @p_g1, @v_6dcircleaxis1_abxad.normalize, @s_len6r, 90
    @pl_6dcircle1 = [@p_g1,@v_6dcircleaxis1_abxad]
#  t_circle1 = self.btcAddText '6dcircle1', @p_g1
    @p_6dcentcircle1 = @p_g1
# Algorithm step 10B - create circle2 centered at g2, radius 6*r
#     Circle will be tangent to ab at e.
    @eds_circle2 = Sketchup.active_model.entities.add_circle @p_g2, @v_6dcircleaxis2_abxad.normalize, @s_len6r, 90
    @p_6dcentcircle2 = @p_g2
#  t_circle2 = self.btcAddText '6dcircle2', @p_g2
    @pl_6dcircle2 = [@p_g2,@v_6dcircleaxis2_abxad]
#-----------------------------------------------------------------
#
#   Middle iteration loop, calls innermost iteration loop, tries each 
#   legitimate segment of circle 1.  "Legitimate" segments are those
#   located on the "above" the plane of face 1, on the side of the bend.
#
#-----------------------------------------------------------------
    i_return = self.btcMiddleIteration
    if i_return != SUCCESS
      return FAILED
    end              
#----------------------------------------------------------------
    @i_edgeidx = @i_edgeidx1
#----------------------------------------------------------------
#   make correction to plane of one circle or the other
#----------------------------------------------------------------
    i_return = self.btcCalculateOOPError
    if i_return != SUCCESS
      return FAILED
    end
    t_ = self.btcAddText @i_adjust.to_s, @p_best1of576
    t_ = self.btcAddText @i_adjust.to_s, @p_best2of576
    p_ = Geom::Point3d.linear_combination 0.5, @p_best1of576, 0.5, @p_best2of576
    t_ = self.btcAddText @i_adjust.to_s, p_
    @i_adjust += 1
    i_show_iteration += 1
#-----------------------------------
   #check for convergence
#-----------------------------------   
    s_ = @s_outofplaneerror
    s_ = s_.abs
    s_ += @s_outofplaneerrorb4
    if s_ * 1000 < 1         # < thousandth of a radian out of plane?
      @i_adjust = @i_adjust_limit  #force loop exit
    end
    s_ += @ss_ooperrors[0]
    s_ += @ss_ooperrors[1]
    @s_oopLast4 = s_
    if @s_oopLast4 > @s_oopLast4b4
      if @s_oopLast4 < 99
        @i_adjust = @i_adjust_limit  #force loop exit    
      end
    end
    @ss_ooperrors[0] = @ss_ooperrors[1]
    @ss_ooperrors[1] = @ss_ooperrors[2]
    @ss_ooperrors[2] = @s_outofplaneerrorb4.abs
    @ss_ooperrors[3] = @s_outofplaneerror.abs
    @s_oopLast4b4 = @s_oopLast4
    s_ = @s_outofplaneerror.abs
    @s_outofplaneerrorb4 = s_
    if $bln_show_construction == 0
      @m_model.entities.erase_entities @eds_circle1
      @m_model.entities.erase_entities @eds_circle2
    end
    t_ = 'Completed iteration ' + i_show_iteration.to_s
    Sketchup::set_status_text  t_, SB_VCB_LABEL
  end  #while @i_adjust<@i_adjust_limit
return SUCCESS
rescue
  self.btcRescuePrint 'btcOutermostIteration'
  return FAILED
end
end
#-----------------------------------------------------------------
def btcCreateBothArcs
begin
  Sketchup::set_status_text 'Starting iterations', SB_VCB_LABEL
  @s_bestofAllTries = 99
  @s_tinner = 0
  i_return = self.btcOutermostIteration
  if i_return != SUCCESS
    return FAILED
  end
  @s_tcalculating = Time.new - @s_tstart
  s_ = @s_tinner / @s_tcalculating
#UI.messagebox('time ratio '+ s_.to_s)
# Algorithm step 13 Replace the circles with an arcs that start and end
#     at the faces and at the tangent points stored.

  i_return = self.btcCreateOneArc 1
  if i_return != SUCCESS
    return FAILED
  end
  i_return = self.btcCreateOneArc 2
  if i_return != SUCCESS
    return FAILED
  end
  return SUCCESS
rescue
  btcRescuePrint 'btcCreateBothArcs'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcAddArcToPath
begin
  eds_arca = @aaa
  p_besta = @bbb
  s_radius_test = @ccc
  @v_arcend = Geom::Vector3d.new [1,2,3]
  if eds_arca == nil
    UI.messagebox('Bend radius is nil')
    return FAILED
  end
  i_limit = eds_arca.size
  i_seg = 0
  ed_edge = eds_arca[i_limit-1]  #scope ed_edge outside its loop            
  while i_seg < i_limit
    ed_edge = eds_arca[i_seg]
    @eds_follow.push ed_edge
    i_seg += 1
  end
  p_last = ed_edge.end.position
  s_diff = p_last.distance p_besta
  if s_diff > 0.1 * s_radius_test
    UI.messagebox('Faces are too close to each other')
    return FAILED
  end
  self.btcAddText('end',p_last)
  p_first=ed_edge.start.position        
  self.btcAddText('start',p_first)
  s_dif = p_last - p_first
  @v_arcend = Geom::Vector3d.new s_dif
  return SUCCESS
rescue
  btcRescuePrint 'btcAddArcToPath'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcConstructContinuousPath
begin
    p_besta = @p_best1of576
    p_bestb = @p_best2of576 
    p_centa = @p_cent1
    p_offa = @p_cent1offset
    p_centb = @p_cent2
    p_offb = @p_cent2offset
    eds_arca = @eds_arc1 
    eds_arcb = @eds_arc2 
    s_radius_test = @s_radius1
  $ed_stub1 = Sketchup.active_model.entities.add_line p_centa, p_offa
  $ed_stub2 = Sketchup.active_model.entities.add_line p_offb, p_centb
  @eds_follow.push $ed_stub1
  $i_stub1ID = $ed_stub1.entityID
  @aaa=eds_arca
  @bbb=p_besta
  @ccc=s_radius_test
  i_return = self.btcAddArcToPath
  if i_return != SUCCESS
    return FAILED
  end
  @v_arcend1 = @v_arcend
  @ed_tangent = Sketchup.active_model.entities.add_line p_besta, p_bestb
  p_end = @ed_tangent.end.position
  p_start = @ed_tangent.start.position
  s_dif = p_end - p_start
  @v_tangent = Geom::Vector3d.new  s_dif
  @eds_follow.push @ed_tangent   #push points from tangent line
  @aaa=eds_arcb
  @bbb=p_bestb
  @ccc=s_radius_test
  i_return = self.btcAddArcToPath
  @v_arcend2 = @v_arcend
  if i_return != SUCCESS
    return FAILED
  end
  s_ang_ends = @v_arcend2.angle_between @v_arcend1
  if s_ang_ends < 160.degrees
    UI.messagebox('FaceS are too close to each other')
    return FAILED
  end
  if s_ang_ends > 200.degrees
    UI.messagebox('Faces Are Too close to each other')
    return FAILED
  end
  s_ang_tan = @v_tangent.angle_between @v_arcend1
  if s_ang_tan > 20.degrees
    if s_ang_tan < 160.degrees
      UI.messagebox('FaceS are too close to Each other')
      return FAILED
    end
  end
  s_ang_tan = @v_tangent.angle_between @v_arcend2.reverse
  if s_ang_tan > 20.degrees
    if s_ang_tan < 160.degrees
      UI.messagebox('FaceS Are too close to each other')
      return FAILED
    end
  end
  @eds_follow.push $ed_stub2
  i_return = self.btcShowPath
  if i_return != SUCCESS
    return FAILED
  end
  return SUCCESS
rescue
  btcRescuePrint 'btcConstructContinuousPath'
  return FAILED
end
end         
#-----------------------------------------------------------------------------
def btcFollowStraightPath(p_1,p_2)
begin
#  if @bln_startAtFace1 == 1 
   
    @ed_tangent = Sketchup.active_model.entities.add_line p_1, p_2
    @eds_follow.push @ed_tangent   #push points from tangent line
    i_return = @f_face1.followme @eds_follow  #use followme tool
                                 #create the outer surface of the extrusion
 # else
 #   @ed_tangent = Sketchup.active_model.entities.add_line p_2,p_1
 #   @eds_follow.push @ed_tangent   #push points from tangent line
 #   i_return = @f_face2.followme @eds_follow
#  end
  return i_return
rescue
  btcRescuePrint 'btcFollowStraightPath'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcSetup
begin
  @i_printed = 0
  @m_model = Sketchup.active_model
  if Sketchup.version[0,1].to_i >= 7
    @m_model.start_operation("draw central path",true)
  else
    @m_model.start_operation "draw central path"
  end
#----------------------------------------------------------------------------------  
  @bln_addtext = 0  #don't need to show entities except for creating documentation
#----------------------------------------------------------------------------------  
  @bln_printingon = 1
  @i_circle_edgecount = 90
  @i_nttcountb4 = Sketchup.active_model.entities.count
  @v_null = Geom::Vector3d.new 0, 0, 0
  @p_origin = Geom::Point3d.new 0, 0, 0
  @v_spare = Geom::Vector3d.new 0, 0, 0
  @eds_follow = []
  @s_btcShowPointradius = 1/25.4
  @s_acceleration_factor = 1 
  @s_6r_criterion = 1.9
  @i_bestcnt = 0
  @s_tstart = Time.new
  @myview = Sketchup.active_model.active_view
  @myview.drawing_color = 'red'
  @bls_usep1 = [1]*@i_circle_edgecount
  @bls_usep2 = [1]*@i_circle_edgecount
  @bls_usep3 = [1]*@i_circle_edgecount
  @bls_usep4 = [1]*@i_circle_edgecount
  @s_angcriterion = ((2*360/@i_circle_edgecount)+1)/57.3  
              #24 segments in circles-->15 degrees, x 2 circles = 30, + 1  gave 30 deg
              #48                                                 15, + 1  gave 10 deg
              #90                                                  8, + 1  gave  4 deg                                                                                                           #48                                             15, +1 gave 10 deg                                                       #90                                             8, +1                                                       
  s_teeny = 0.000000000000001
  @v_teeny = Geom::Vector3d.new s_teeny, s_teeny, s_teeny
  return SUCCESS
rescue Exception  >= t_exception
  btcRescuePrint 'EXCEPTION IN btcSetup'
  puts 't_exception = '+t_exception.to_s
  @m_model.commit_operation
  return FAILED
rescue
  btcRescuePrint 'btcSetup'
  @m_model.commit_operation
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcTestReversal(p_a,p_b,i_1or2)
begin
  s_d = p_a.distance p_b
  if s_d < 0.1
    if i_1or2 == 1 then
      $bln_reverse_stub1 = 1
    else
      $bln_reverse_stub2 = 1
    end
  end
  return s_d
rescue
  btcRescuePrint 'btcTestReversal'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcFindEndSelector1(e_)
begin
  @p_locator1 = nil
  if e_.typename == 'Edge'
    @p_locator1 = e_.start.position
  end
  if e_.typename == 'Vertex'
    @p_locator1 = e_.position
  end
  return SUCCESS
rescue
  btcRescuePrint 'FindEndSelector1'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcFindEndSelector2(e_)
begin
  @p_locator2 = nil
  if e_.typename == 'Edge'
    @p_locator2 = e_.start.position
  end
  if e_.typename == 'Vertex'
    @p_locator2 = e_.position
  end
  return SUCCESS
rescue
  btcRescuePrint 'FindEndSelector2'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcGetPreselectedFaces
begin
  @fs_connected2 = nil
  @p_locator1 = nil
  @p_locator2 = nil
  $i_face_count = 0
  bln_have_face1 = 0
  $bln_reverse_stub1 = 0
  $bln_reverse_stub2 = 0
  selection = Sketchup.active_model.selection
  selection.each { |entity| 
    $bln_found_face1 = 0
    if entity.typename != 'Face'
      if entity.typename == 'Edge'
        ed_ = entity
        @fs_connected2 = ed_.faces
        if bln_have_face1 == 0
          @fs_connected = ed_.faces
          if @fs_connected != nil
            if @fs_connected.length > 0
              bln_have_face1 = 1
              $bln_found_face1 = 1
            end
          else  
          end
        end
      end 
      if $bln_found_face1 == 0
        if (@fs_connected2 == nil) || (@fs_connected2.length == 0)
          if @p_locator1 == nil
            i_return = self.btcFindEndSelector1(entity) 
            if i_return != SUCCESS
              return FAILED 
            end
          else
            if @p_locator2 == nil
              i_return = self.btcFindEndSelector2(entity)
              if i_return != SUCCESS
                return FAILED
              end 
            end
          end  
        end
      end
    else
      $i_face_count += 1
      if $i_face_count == 1
        @f_face1 = entity
      end
      if $i_face_count == 2
        @f_face2 = entity
      end
      if $i_face_count > 2
#        UI.messagebox('Error > 2 faces selected.  Using first 2.')
#        return FAILED
        $i_face_count = 2
      end
    end
    }
  if $i_face_count == 0
    UI.messagebox('0 faces selected.')
    return FAILED
  end
  if $i_face_count == 1
    UI.messagebox('Only 1 face selected.')
    return FAILED
  end
  if bln_have_face1 == 1
    i_return = self.btcInterchangeFaces
    if i_return != SUCCESS
      return FAILED
    end
  end
  i_return = btcProcessFace1  #do this the first time just to calculate the face center
  if i_return != SUCCESS
    return FAILED
  end
  i_return = btcProcessFace2#do this the first time just to calculate the face center
  if i_return != SUCCESS
    return FAILED
  end
  if @p_locator1 != nil

if @p_cent1 == nil
else
end

    s_dis1 = @p_locator1.distance @p_cent1
    s_dis2 = @p_locator1.distance @p_cent2
    if s_dis2 < s_dis1
      $bln_reverse_stub2 = 1
    else
      $bln_reverse_stub1 = 1
    end
  end
  if @p_locator2 != nil
    s_dis1 = @p_locator2.distance @p_cent1
    s_dis2 = @p_locator2.distance @p_cent2
    if s_dis2 < s_dis1
      $bln_reverse_stub2 = 1
    else
      $bln_reverse_stub1 = 1
    end
  end
  i_return = btcProcessFace1
  if i_return != SUCCESS
    return FAILED
  end
  i_return = btcProcessFace2
  return i_return
rescue
  btcRescuePrint 'btcGetPreselectedFaces'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcInterchangeFaces
begin
    i_return = self.btcCalc2Centroids
    if i_return != SUCCESS
      return FAILED
    end
    @e_myentity = @fs_connected[0]
    i_return = self.btcCalculateFaceCentroid
    if i_return != SUCCESS
      return FAILED
    end
    p_r_ind1 = @p_cent
    s_to_1 = p_r_ind1.distance @p_cent1
    s_to_2 = p_r_ind1.distance @p_cent2
    if s_to_2 < s_to_1
      f_ = @f_face2
      @f_face2 = @f_face1
      @f_face1 = f_
    end
  return SUCCESS
rescue
  btcRescuePrint 'btcInterchangeFaces'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcCalc2Centroids
begin
    @e_myentity = @f_face1
    i_return = self.btcCalculateFaceCentroid
    if i_return != SUCCESS
      return FAILED
    end
    @p_cent1 = @p_cent
    @e_myentity = @f_face2
    i_return = self.btcCalculateFaceCentroid
    if i_return != SUCCESS
      return FAILED
    end
    @p_cent2 = @p_cent
  return SUCCESS
rescue
  btcRescuePrint 'btcCalc2Centroids'
  return FAILED
end
end

#-----------------------------------------------------------------------------
def btcProcessFace1
begin
  @m_model.selection.add @f_face1
  @e_myentity = @f_face1
  i_return = self.btcMakev_unitnormal
  if i_return != SUCCESS
    return FAILED
  end
  @p_cent1 = @p_cent
  @s_radius1 = @s_radius
  @v_unitnormal1 = @v_unitnormal
  return SUCCESS
rescue Exception  >= t_exception
  btcRescuePrint 'EXCEPTION IN processface1'
  puts 't_exception = '+t_exception.to_s
  @m_model.commit_operation
rescue
  btcRescuePrint 'processface1'
  @m_model.commit_operation
  return 2
end
end
#-----------------------------------------------------------------------------
def btcProcessFace2
begin
  @m_model.selection.add @f_face2
  @e_myentity = @f_face2
  i_return = self.btcMakev_unitnormal
  if i_return != SUCCESS
    return FAILED
  end
  @p_cent2 = @p_cent
  @s_radius2 = @s_radius
  @v_unitnormal2 = @v_unitnormal
  return SUCCESS
rescue Exception  >= t_exception
  btcRescuePrint 'EXCEPTION IN processface2'
  puts 't_exception = '+t_exception.to_s
  @m_model.commit_operation
rescue
  btcRescuePrint 'processface2'
  @m_model.commit_operation
  return 2
end
end
#-----------------------------------------------------------------------------
def btcShowPoint(p_x,bln_pause)
begin
  if $bln_show_points == 0
    return SUCCESS
  end
  @v_spare.x = 1
  @v_spare.y = 0
  @v_spare.z = 0
  s_len = @s_btcShowPointradius
  eds_cirkle1 = Sketchup.active_model.entities.add_circle p_x, @v_spare, s_len, @i_circle_edgecount
  @v_spare.x = 0
  @v_spare.y = 1
  eds_cirkle2 = Sketchup.active_model.entities.add_circle p_x, @v_spare, s_len, @i_circle_edgecount  

  @v_spare.y = 0
  @v_spare.z = 1
  eds_cirkle3 = Sketchup.active_model.entities.add_circle p_x, @v_spare, s_len, @i_circle_edgecount  

  if bln_pause != 0
    UI.messagebox('Showing point')
  end
  return SUCCESS
rescue
  btcRescuePrint 'btcShowPoint'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcShowPath
begin
  i_count = @eds_follow.length
  i_idx = 0
  while i_idx < i_count
    e_ = @eds_follow[i_idx]
    l_line = e_.line
    p_st = l_line[0]
    i_return = self.btcShowPoint p_st,0
    if i_return != SUCCESS
      return FAILED
    end
    i_idx += 1
  end
  return SUCCESS
rescue
  btcRescuePrint 'btcShowPath'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcMakeSpecialDistance
begin
  p_temp = [@v_unitnormal1.x, @v_unitnormal1.y,  @v_unitnormal1.z]
  t_=p_temp.to_s
  t_2 = '@p_cent1 '+@p_cent1.to_s
#UI.messagebox(t_+t_2)
  @p_unitend1 = @p_cent1 + p_temp
                    #contruct point one inch out from centroid 1 
  tr_ = Geom::Transformation.scaling @s_len6r
  v_6r = @v_unitnormal1.transform tr_
  @p_6rnormend1 = @p_cent1 + v_6r
  p_temp = [@v_unitnormal2.x, @v_unitnormal2.y,  @v_unitnormal2.z]
  @p_unitend2 = @p_cent2 + p_temp 
  v_6r = @v_unitnormal2.transform tr_
  @p_6rnormend2 = @p_cent2 + v_6r
  @s_len12 = @p_6rnormend1.distance @p_6rnormend2
  return SUCCESS
rescue
  btcRescuePrint 'btcMakeSpecialDistance'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def btcGetBestNormals
begin
#try four outward normal direction possibilities, use the combination with 
#  minimum separation of the end points
  if self.btcMakeSpecialDistance  != SUCCESS       #combo a - original normals
    return FAILED
  end
  s_bestlen = @s_len12                 #best so far = minimum 
  i_bestdir1 = 1
  i_bestdir2 = 1
  @v_unitnormal1 = @v_unitnormal1.reverse   #reverse one normal
  if self.btcMakeSpecialDistance != SUCCESS           #combo b
    return FAILED
  end
  if @s_len12<s_bestlen
    i_bestdir1 = -1
    i_bestdir2 = 1
    s_bestlen = @s_len12
  end
  @v_unitnormal1 = @v_unitnormal1.reverse   #put normal 1 back the way it was
  @v_unitnormal2 = @v_unitnormal2.reverse   #reverse the other one     combo c
  if self.btcMakeSpecialDistance  != SUCCESS
    return FAILED
  end
  if @s_len12<s_bestlen
    i_bestdir1 = 1
    i_bestdir2 = -1
    s_bestlen = @s_len12
  end
  @v_unitnormal1 = @v_unitnormal1.reverse   #now both are reversed    combo d
  if self.btcMakeSpecialDistance != SUCCESS 
    return FAILED
  end
  if @s_len12<s_bestlen
    i_bestdir1 = -1
    i_bestdir2 = -1
    s_bestlen = @s_len12
  end
#make this easier to understand by getting back to the original normals 
  @v_unitnormal1 = @v_unitnormal1.reverse   #put normal 1 back the way it was
  @v_unitnormal2 = @v_unitnormal2.reverse   #put normal 2 back the way it was

#use normal directions giving closest distance to end points
  if i_bestdir1 < 0
    @v_unitnormal1 = @v_unitnormal1.reverse
  end
  if i_bestdir2 < 0
    @v_unitnormal2 = @v_unitnormal2.reverse
  end
  if $bln_reverse_stub1 == 1

    @v_unitnormal1 = @v_unitnormal1.reverse
  end
  if $bln_reverse_stub2 == 1
    @v_unitnormal2 = @v_unitnormal2.reverse
  end
#recalculate with new directions
  if self.btcMakeSpecialDistance  != SUCCESS            #use s_best combo
    return FAILED
  end
  return SUCCESS
rescue
  btcRescuePrint 'btcGetBestNormals'
  return FAILED
end
end
#-----------------------------------------------------------------------------
def activate
begin
  if $i_face_count == 0
    i_return = self.btcSetup
    if i_return != SUCCESS
      Sketchup::set_status_text 'F2F tool failed 1', SB_VCB_LABEL
      self.btcGiveUpControl
      return FAILED
    end
    Sketchup::set_status_text 'F2F setup ok', SB_VCB_LABEL
    i_return = self.btcGetPreselectedFaces
    if i_return != SUCCESS
      Sketchup::set_status_text 'F2F tool failed 2', SB_VCB_LABEL
      self.btcGiveUpControl
      return FAILED
    end
  end
  if $i_face_count == 2
    i_return = self.btcMakeExtrusion
    if i_return != SUCCESS
      Sketchup::set_status_text 'F2F tool failed 5', SB_VCB_LABEL

    else
      Sketchup::set_status_text 'F2F tool succeeded', SB_VCB_LABEL
    end
    if $bln_label_faces == 1
      t_ = @m_model.entities.add_text 'face1', @p_cent1
      t_ = t_.to_s
      t_ = @m_model.entities.add_text 'face2', @p_cent2
      t_ = t_.to_s
      bln_refreshed_view = @m_model.active_view.refresh 
    end    
    self.btcGiveUpControl
    return i_return
  end
  return FAILED
rescue
  btcRescuePrint 'activate'
  return FAILED
end
end
#-----------------------------------------------------------------------------
end # class Face2Face
#-----------------------------------------------------------------------------
def MF2F.tool()
# Select the Face2Face tool
    Sketchup.active_model.select_tool Face2Face.new()
end
#-----------------------------------------------------------------------------
def MF2F.parameterstool()
begin
  ts_Prompts=["Bend radius multiplier","Extrusion radius multiplier","Stub length multiplier"]
#  ts_Prompts.push "Reverse direction of last extrusion "
  ts_Prompts.push "Show contruction"
  ts_Prompts.push "Show points"
  ts_Prompts.push "Animate search for best solution"
  ts_Prompts.push "Label faces "
  t_1 = $s_bend_radius_multiplier.to_s
  t_2 = $s_extr_rad_multiplier.to_s
 # ts_Defaults=[t_1,t_2,"1","0"]
  ts_Defaults=[t_1,t_2,"1"]
  ts_Defaults.push $bln_show_construction.to_s
  ts_Defaults.push $bln_show_points.to_s
  ts_Defaults.push $bln_animate.to_s
  ts_Defaults.push $bln_label_faces.to_s
  t_=''
  ts_input=UI.inputbox(ts_Prompts,ts_Defaults,t_)
  bln_bad = 0                            
  s_ = ts_input[0].to_f
  if s_ > 0.1
    if s_ < 100
      s_0 = s_
    else
      UI.messagebox('Bad value for bend radius multiplier, not changed')
      bln_bad = 1
    end
  end
  s_ = ts_input[1].to_f
  if s_ > 0.1
    if s_ < 100
      s_1 = s_
    else
      UI.messagebox('Bad value for extrusion radius multiplier, not changed')
      bln_bad = 1
    end
  end
  s_ = ts_input[2].to_f
  if s_ > 0.1
    if s_ < 100
      s_2 = s_
    else
      UI.messagebox('Bad value for stub length multiplier, not changed')
      bln_bad = 1
    end
  end
 # i_ = ts_input[3].to_i
 # if i_ == 1
 #     i_3 = i_
 # elsif i_ == 0
 #     i_3 = i_
 # else
 #   UI.messagebox('Bad value for reversal, not changed')
 ##   bln_bad = 1
 # end
  i_ = ts_input[3].to_i
  if i_ == 1
    i_4 = i_
  elsif i_ == 0
    i_4 = i_
  else
    UI.messagebox('Bad value for construction, not changed')
    bln_bad = 1
  end
  i_ = ts_input[4].to_i
  if i_ == 1
    i_5 = i_
  elsif i_ == 0
    i_5 = i_
  else
    UI.messagebox('Bad value for points, not changed')
    bln_bad = 1
  end
  i_ = ts_input[5].to_i
  if i_ == 1
    i_6 = i_ 
  elsif i_ == 0
    i_6 = i_ 
  else
    UI.messagebox('Bad value for animation, not changed')
    bln_bad = 1
  end
  i_ = ts_input[6].to_i
  if i_ == 1
    i_7 = i_ 
  elsif i_ == 0
    i_7 = i_ 
  else
    UI.messagebox('Bad value for labeling faces, not changed')
    bln_bad = 1
  end
  if bln_bad == 0 
    $s_bend_radius_multiplier = s_0
    $s_extr_rad_multiplier = s_1
    $s_stub_length_multiplier = s_2
    $bln_show_construction = i_4
    $bln_show_points = i_5
    $bln_animate = i_6
    $bln_label_faces = i_7
  end
  if $bln_reverse_followpath == 1
    i_return = Face2Face.activate
  end
  return SUCCESS
rescue
  Face2Face.btcRescuePrint 'parameterstool'
  return FAILED
end
end
#-----------------------------------------------------------------------------
end # module MF2F
#-----------------------------------------------------------------------------
# Add a menu choice for creating extrusion
  unless file_loaded?('Face2Face.rb')
    file_loaded('Face2Face.rb')
    $submenu ? (organizer = $submenu) : (organizer = UI.menu('Plugins'))
    organizer.add_item('Extrude Face2Face') { MF2F.tool }
    organizer.add_item('Face2Face Parameters') { MF2F.parameterstool }
  end
#end
#-----------------------------------------------------------------------------
